// Copyright - Simovits Consulting 2009
// Author: James Dickson.


#include <winsock2.h>
#include <direct.h>
#include <string>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/applink.c> // Needed for PEM_ write functions

#include "common.h"
#include "sslserver.h"
#include "argparser.h"
#include "wincolor.h"


#define PROTOCOL_COUNT 3
#define PROTOCOL_TESTS	4
#define MAX_CIPHERITEMS 1024

using namespace std;

// Prototypes
int doTest(TESTINFO testInfo, 
		   char *strCipher, 
		   RequestProtocol protocol, 
		   CERTINFO *certInfo, 
		   long *certRes, 
		   int *algBits, 
		   int *actuallyUsed, 
		   BOOL *bRenegotiatePossible,
		   BOOL *bSafeRenegotiation);
int sendRequest(SSL *ssl, TESTINFO testInfo);
int receiveResponse(SSL  *ssl, TESTINFO testInfo);
int check_cert(SSL *ssl, char *host, X509 *peer, long *certResult);

// Global variables
char strPemDump[1024];
X509 *crtCurrent;
BOOL bCertTested;
//BOOL bCertIsSelfSigned;

/**
 * Init the Win32 socket
 */
int initWSA()
{
#ifdef WIN32
	WSADATA wsaData;

	if(WSAStartup(MAKEWORD(2,0),&wsaData) != 0)
	{
		return 0;
	}

#endif
	
	return 1;
}




/**
 * return the protocol description.
 */
char * getProtocolDescription(RequestProtocol rq)
{
	if(rq == SSL2) return "SSL 2";
	if(rq == SSL3) return "SSL 3";
	if(rq == TLS1) return "TLS 1";
	if(rq == DTLS1) return "DTLS 1";

	return "";
}


/**
 * Init libraries
 */
int initLibraries()
{
	int initSocket = initWSA();	// Initialize winsock

	// Check the WSA-init result.
	if(initSocket != 1)
	{
		printf("[-] ERROR: Could not initialize Win32 socket!\n");
		exit(-1);
	}
	
	strcpy(strPemDump, "");	// Set Pem-dump-file to zero length
	bCertTested = FALSE;	// Cert is not tested from start.
	//bCertIsSelfSigned = FALSE; // Give the certificate a chance.

	// Init SSL
    SSL_library_init();
    SSL_load_error_strings();	

	return 0;
}

/**
 * Main procedure.
 */
int main(int argc, char **argv)
{
	TESTINFO testInfo;	
	RequestProtocol protocols[PROTOCOL_TESTS] = { SSL2, SSL3, TLS1, DTLS1};
	FOUNDSUCCESS foundSuccess[1024];
	int successCount = 0;
	int lowCount = 0;
	int highCount = 0;
	int mediumCount = 0;
	BOOL chosenProtocols[4] = { FALSE, FALSE, FALSE, FALSE };
	int certErrors = 0;


	int bestKeyLength = -10;		// For Key Exchange.

	char strCertType[256];			// Will contain the cert. type.
	BOOL bHasAnonymous = FALSE;		// If true then we have an anonymous-keyexchange
	BOOL bHasExport = FALSE;		// If true then we have an export-keyexchange.
	CIPHERTESTITEM cItems[MAX_CIPHERITEMS];
	int actuallyTested = 0;			// Will contain how many ciphers that were tested.
	BOOL bRenegotiationPossible = FALSE;	// If true then we can renegotiate after connection.
	BOOL bSecureRenegotiationPossible = FALSE;		// If true then secure renegotiation is supported
	BOOL bSignSHA1 = FALSE;
	BOOL bSignMD5 = FALSE;
	BOOL bSignMD2 = FALSE;
	BOOL bSignMD4 = FALSE;
	BOOL bGotCert = FALSE;
	int certType = -1;				// Will contain certificate type.

	
	// Init win32 console functions
	initConsole();

	setColor(COLOR_BLUISH);

	// Write banner
	printf("Simovits Consulting - SSL Diagnostics tool v0.6.9a - www.simovits.com\n\n");

	restoreColor();

	initTestInfo(&testInfo);	// Just initialize the struct.
	initLibraries();			// Init wsock, ssl and globals
	


	// Parse the arguments
	int pret = parseArguments(argc, argv, &testInfo, strPemDump, NULL, chosenProtocols);

	// Configure color usage.
	enableColors(testInfo.bUseColors);

	// Get the available ciphers.
	int availableCipherCount = getAvailableCiphers(cItems, MAX_CIPHERITEMS, testInfo);

	if(testInfo.bVerbose)
	{
		printf("[+] Available cipher/protocol combinations: %d\n\n", availableCipherCount);
	}

	if(testInfo.bPrintAvailableCiphers)
	{
		printf("[+] Available ciphers:\n");
		for(int x=0; x < availableCipherCount; x++)
		{
			int strength = cItems[x].cinf.strength;

			if(testInfo.bVerbose)
			{
				printf("(%2d) : %s : %s : %s : %d : %s \n", 
					(x+1), 
					getProtocolDescription(cItems[x].protocol), 
					cItems[x].cinf.strOpenSSLShortName,
					cItems[x].cinf.strOpenSSLName,
					cItems[x].cinf.bits,
					cItems[x].cinf.strStrengthName
					);
			}
			else
			{
				printf("(%2d) : %s : %s : %d : %s \n", 
					(x+1), 
					getProtocolDescription(cItems[x].protocol), 
					cItems[x].cinf.strOpenSSLShortName,
					cItems[x].cinf.bits,
					cItems[x].cinf.strStrengthName
					);
			}
		}
	}
	
	// Check for valid IP-adress
	if(testInfo.strIPAdress == NULL && !testInfo.bServerMode) 
	{
		if(!testInfo.bPrintAvailableCiphers)
		{
			printf("[-] You must specify a valid target! \n");

			if(testInfo.strHostname == NULL)
			{
				printUsage();
			}
			else
			{
				printf("[-] The target %s is not valid.", testInfo.strHostname);
			}

			return -1;
		}

		return 0;
	}

	if(pret != 2)
	{
		// If not valid then we return.
		return pret;
	}

	// Print user supplied info
	if(!testInfo.bSuppressUserSupplied)
	{
		printf("----------------TEST INFO---------------------------\n");
		printf("[*] Target IP:        %s\n", testInfo.strIPAdress);
		printf("[*] Target Hostname:  %s\n", testInfo.strHostname);
		printf("[*] Target Port:      %d\n", testInfo.port);

		printf("----------------------------------------------------");
		printf("\n\n");
	}




	// Server mode
	if(testInfo.bServerMode)
	{
		printf("[+] Running in server mode...\n");

		if(testInfo.strServerCipher == NULL)
		{
			setColor(COLOR_YELLOW);
			printf("[!] WARNING: No specific ciphers specified reverting to ALL!\n");
			restoreColor();

			testInfo.strServerCipher = "ALL:COMPLEMENTOFALL";
		}
		printf("    [*] Protocol: %s\n"
			   "    [*] Cipher: %s\n"
			   "    [*] Server certificate: %s\n", 
			   getProtocolDescription(testInfo.serverProtocol), 
			   testInfo.strServerCipher,
			   testInfo.strServerCertPath);

		serverAcceptLoop(testInfo, testInfo.serverProtocol);
		return 0;
	}


	// ----------- Tests starts here --------------------
	// Now we do the tests		
	for(int i=0; i < availableCipherCount; i++)
	{
		long certRes = 0;
		CRYPTO_INFO cinfo = cItems[i].cinf; // 2010-07-07 - james

		// If the protocol is not chosen then go to next.
		if(testInfo.bTestOnlySpecificProtocols)
		{			 
			if(chosenProtocols[cItems[i].protocol] != TRUE)
			{
				continue;
			}
		}

		// If we have not chosen UDP-connection and we are at the "DTLS1"-test, then it is meaningless
		// to test for DTLS1
		if((! testInfo.bUDP) && (cItems[i].protocol == DTLS1))
		{
			continue;
		}

		// If we have chosen UDP-connections but the protocol is not DTLS1, then it is meaningless to do the test.
		if(testInfo.bUDP && (cItems[i].protocol != DTLS1))
		{
			continue;
		}


		char cryptoGrade[256];	// Will contain the name of the grade (low, medium, high)
		int grade = cItems[i].cinf.strength;

		// Check for the grade.
		if(testInfo.onlyGrade >= 0)
		{
			if(grade != testInfo.onlyGrade)
			{
				continue;
			}
		}
		
		actuallyTested++;					// Increase counter.
		cItems[i].testCount++;				// Increase test counter.


		// Print out the trace information.
		printf("[+] Testing %s: %s, %s (%d bits) ... ", cItems[i].cinf.strStrengthName , getProtocolDescription(cItems[i].protocol), cinfo.strOpenSSLShortName, cinfo.bits);

		foundSuccess[successCount].cinfo = cItems[i];

		// Perform the test.
		int algBits = 0;
		int actuallyUsed = 0;
		BOOL bRenegotiated = FALSE;
		BOOL bSafeRenegotiation = FALSE;
		int result = doTest(testInfo, 
							cinfo.strOpenSSLShortName, 
							cItems[i].protocol, 
							&foundSuccess[successCount].certInfo, 
							&certRes, 
							&algBits, 
							&actuallyUsed,
							& bRenegotiated,
							&bSafeRenegotiation);
		

		// If we could connect.
		if(result >= 1)
		{						
			foundSuccess[successCount].bConnect =	TRUE;
			foundSuccess[successCount].bReceive =	FALSE;
			foundSuccess[successCount].bSend  =		FALSE;
			foundSuccess[successCount].bRenegotiate = bRenegotiated;
			foundSuccess[successCount].bSafeRenegotiate = bSafeRenegotiation;

			if(bRenegotiated)
			{
				bRenegotiationPossible = TRUE;
			}

			if(bSafeRenegotiation)
			{
				bSecureRenegotiationPossible = TRUE;
			}
			

			// Here we make the assumption that the export-cryptos are 40 and Anonymous 0 bits.
			if(cinfo.bits == 0)
			{
				bHasAnonymous = TRUE;
			}
			else if(cinfo.bits == 40)
			{
				bHasExport = TRUE;
			}

			// Check certificate
			if(foundSuccess[successCount].certInfo.certficate)
			{
				bGotCert = TRUE;
				certType = foundSuccess[successCount].certInfo.certficate->cert_info->key->pkey->type;

				const int ALG_BUFLEN = 1024;
				char strCertBuf[ALG_BUFLEN];
				if(getCertAlgortithm(foundSuccess[successCount].certInfo.certficate, strCertBuf, ALG_BUFLEN))
				{				
					// Convert to upper case
					strtoupper(strCertBuf, ALG_BUFLEN);

					char *strMD5 = strstr(strCertBuf, "MD5");
					char *strMD2 = strstr(strCertBuf, "MD2");
					char *strSHA1 = strstr(strCertBuf, "SHA1");
					char *strMD4 = strstr(strCertBuf, "MD4");

					if(strMD5) bSignMD5 = TRUE;
					if(strMD4) bSignMD4 = TRUE;
					if(strMD2) bSignMD2 = TRUE;
					if(strSHA1) bSignSHA1 = TRUE;					
				}
			}

			// Get the key length of the certificate.
			if(testInfo.bCheckCert)
			{
				X509 * tcert = foundSuccess[successCount].certInfo.certficate;

				if(tcert != NULL)
				{
					int certBitCount =  getCertBitCount(tcert);

					if(bestKeyLength < certBitCount)
					{
						bestKeyLength = certBitCount;
					}
					

					// Get the certificate type description
					getCertTypeString(tcert->cert_info->key->pkey->type, strCertType);
				}
			}

			// Count grades.
			if(grade <= 1)
			{
				lowCount++;
			}
			else if(grade == 2)
			{
				mediumCount++;
			}
			else if(grade == 3)
			{
				highCount++;
			}

			// Print status
			if( (result & CONNECT_OK) > 0)
			{
				// Set color.
				if(grade == 1)
				{
					setColor(COLOR_RED);
				}
				else if(grade == 2)
				{
					setColor(COLOR_YELLOW);
				}
				else
				{
					setColor(COLOR_GREEN);
				}

				// Print ok. message.
				printf("CONNECT_OK ");

				// Restore color
				restoreColor();
			}
			else
			{
				printf("CONNECT_FAIL ");
			}

			// If we should test HTTP
			if(testInfo.bTestSend)
			{
				if((result & SEND_OK) > 0)
				{
					printf("SEND_OK ");
					foundSuccess[successCount].bSend = TRUE;
				}
				else
				{
					printf("SEND_FAIL ");
				}


				if((result & RECEIVE_OK) > 0)
				{
					printf("RECEIVE_OK ");
					foundSuccess[successCount].bReceive = TRUE;
				}
				else
				{
					printf("RECEIVE_FAIL ");
				}
			}

			// Check the certificate...
			if(certRes == X509_V_OK)
			{
				printf("CERT_OK ");
			}
			else
			{
				printf("\n[-] CERT_ERROR: %s \n", X509_verify_cert_error_string(certRes));

				if((certRes & CERT_ERROR_CN_MISMATCH) > 0)
				{
					printf("[-] CERT_ERROR: Common name mismatch\n");
				}

				certErrors++;
			}

			successCount++;
			printf("\n\n");
		}
		else 
		{
			printf("FAILED\n");
		}
		
		// Sleep before next test...
		if(testInfo.msInterval > 0)
		{
			Sleep(testInfo.msInterval);
		}				
	}

	printf("\n");


	// Show summary
	if(! testInfo.bSuppressSummary)
	{
		setColor(COLOR_BLUISH);
		printf("\n\n-= SUMMARY =-\n\n");
		restoreColor();

		printf("Weak:         %d\n", lowCount);
		printf("Intermediate: %d\n", mediumCount);
		printf("Strong:       %d\n", highCount);

		// 2010-11-05
		if(testInfo.bPrintCert)
		{
			printCert(crtCurrent);
		}


		// How many were tested...
		if(testInfo.bVerbose)
		{
			printf("[+] Available cipher-protocol-combinations: %d were %d were tested\n\n", availableCipherCount, actuallyTested);

			if(availableCipherCount > actuallyTested)
			{
				printf("[+] The following were not tested:\n");

				for(int x=0; x < availableCipherCount; x++)
				{
					if(! cItems[x].testCount <= 0)
					{
						printf("  * %s - %s - %d\n", getProtocolDescription(cItems[x].protocol), cItems[x].cinf.strOpenSSLShortName, cItems[x].testCount);
					}
				}
			}
		}

		// Check for certificate bit length
		if(	bGotCert && 
			bestKeyLength < 1024 && 
			(certType == EVP_PKEY_RSA || certType == EVP_PKEY_DSA))
		{
			setColor(COLOR_RED);
			printf("\n[!] WARNING!: Certificate key length is less than recommended for RSA and DSA (%d)\n\n", bestKeyLength);
			restoreColor();
		}

		// Check for certificate bit length for Elliptic Curve
		if(	bGotCert && 
			bestKeyLength < 160 && 
			certType == EVP_PKEY_EC)
		{
			setColor(COLOR_RED);
			printf("\n[!] WARNING!: Certificate key length is less than recommended for EC (%d)\n\n", bestKeyLength);
			restoreColor();
		}

		// Check for certificate sign algorithm
		if(bSignMD2)
		{
			setColor(COLOR_RED);
			printf("\n[!] WARNING!: Certificate is signed with MD2 algorithm\n\n");
			restoreColor();
		}

		if(bSignMD4)
		{
			setColor(COLOR_RED);
			printf("\n[!] WARNING!: Certificate is signed with MD4 algorithm\n\n");
			restoreColor();
		}

		if(bSignMD5)
		{
			setColor(COLOR_RED);
			printf("\n[!] WARNING!: Certificate is signed with MD5 algorithm\n\n");
			restoreColor();
		}

		if(bSignSHA1)
		{
			setColor(COLOR_YELLOW);
			printf("\n[!] WARNING!: Certificate is signed with SHA1 algorithm\n\n");
			restoreColor();
		}

		// Check the certificate errors...
		if(certErrors > 0)
		{
			setColor(COLOR_RED);
			printf("\n[!] WARNING!: Certificate errors exists.\n\n");
			restoreColor();
		}

		// Check for renegotiation
		if(bRenegotiationPossible)
		{
			if(bSecureRenegotiationPossible)
			{
				setColor(COLOR_YELLOW);
				printf("\n[!] Warning: Renegotiation possible. Secure renegotiation is however supported. \n\n");
				restoreColor();
			}
			else
			{
				setColor(COLOR_RED);
				printf("\n[!] WARNING!: Renegotiation possible and secure renegotiation is Not supported. \n\n");
				restoreColor();
			}
		}
		else if(testInfo.bVerbose)
		{
			printf("\n[!] Information: Renegotiation is NOT possible.\n");
		}

		if((lowCount + mediumCount) > 0)
		{
			setColor(COLOR_RED);
			printf("\n[!] WARNING!: Cipher/methods-combinations not considered strong are enabled.\n\n");
			restoreColor();

			printf("[!] The following non-strong ciphers/methods are enabled at the host:\n\n");

			for(int i=0; i < successCount; i++)
			{
				if(foundSuccess[i].cinfo.cinf.strength <= 2)
				{
					printf("\n");
					printf("[+] Protocol:    %s\n", getProtocolDescription(foundSuccess[i].cinfo.protocol));
					printf("    Cipher name: %s\n", foundSuccess[i].cinfo.cinf.strOpenSSLShortName);
					printf("    Key bits:    %d\n", foundSuccess[i].cinfo.cinf.bits);

					if(foundSuccess[i].cinfo.cinf.strength == 1) setColor(COLOR_RED);
					if(foundSuccess[i].cinfo.cinf.strength == 2) setColor(COLOR_YELLOW);

					printf("    Grade:       %s\n", foundSuccess[i].cinfo.cinf.strStrengthName);

					restoreColor();

					//printf("Grade: %d\n", foundSuccess[i].grade);

					if(foundSuccess[i].bConnect)
					{
						printf("    Connect:     OK\n");

						if(testInfo.bTestSend)
						{
							if(foundSuccess[i].bSend)
							{
								printf("    Send:        OK\n");
							}
							else
							{
								printf("    Send:        FAIL\n");
							}

							if(foundSuccess[i].bReceive)
							{
								printf("    Receive:     OK\n");
							}
							else
							{
								printf("    Receive:     FAIL\n");
							}

						}
					}
				}
			}
		}
	}
    
	WSACleanup(); // Cleanup Win32 socket

	return 0;
}


/**
 * Check certificate.
 */
int check_cert(SSL *ssl, char *host, X509 *peer, SSL_CTX *ctx, long *certResult)
{
	//char peer_CN[256];
	int retval = 0;    

	// Check the certificate
	peer=SSL_get_peer_certificate(ssl);

	//STACK_OF(X509) *chain = SSL_get_peer_cert_chain(ssl);
	

	// certdate = X509_get_notAfter(connssl->server_cert);
	// issuer = PEM_read_X509(fp,NULL,ZERO_NULL,NULL);
	// if (X509_check_issued(issuer,connssl->server_cert) != X509_V_OK) 
	

	// Add the CRL.
	X509_LOOKUP *lookup=X509_STORE_add_lookup(ctx->cert_store, X509_LOOKUP_file());

	if(lookup)
	{
		X509_STORE_set_flags(ctx->cert_store, X509_V_FLAG_CRL_CHECK | X509_V_FLAG_CRL_CHECK_ALL);
	}
	else
	{
		setColor(COLOR_YELLOW);
		printf("[!] WARNING: Cannot load CRL!\n");
		restoreColor();
	}

	// Get the result
	*certResult = SSL_get_verify_result(ssl);

	// Verify the certficate (always fails for SSL v2.0).
    if(*certResult != X509_V_OK)
	{
		retval |= CERT_ERROR_NOT_VERIFIED;
	}

	// Get the Common Name
    //X509_NAME_get_text_by_NID(X509_get_subject_name(peer), NID_commonName, peer_CN, 256);
	//X509_NAME_get_text_by_NID(X509_get_subject_name(peer), NID_host, peer_CN, 256);
	//X509_NAME_get_text_by_NID(X509_get_subject_name(peer), NID_issuing_distribution_point, peer_CN, 256);
	//X509_NAME_get_text_by_NID(X509_get_subject_name(peer), NID_dnQualifier, peer_CN, 256);
	
	/*
	
	
    // Compare host name with the host.
	if(strcmpi(peer_CN, host) != 0)
	{
		retval |= CERT_ERROR_CN_MISMATCH;
	}
	*/

	return retval;
}



/**
 * Verify certificate callback routine.
 */
static int cert_verify_callback(int ok, X509_STORE_CTX *ctx)
{
	// If not cert. is not tested yet
	if(! bCertTested)
	{
		X509 *crt = X509_STORE_CTX_get_current_cert(ctx);		// Get the cert	

		/*
		int cError = X509_STORE_CTX_get_error(ctx);

		if(cError == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT)
		{
			bCertIsSelfSigned = TRUE;
		}
		*/


		// Save ptr to the current certificate
		crtCurrent = crt;		

		// If we should store the .cer file.
		if(strlen(strPemDump) > 0)
		{	  
			BIO *f = BIO_new_file(strPemDump, "w"); 	// Create new file		 
			PEM_write_bio_X509(f, crt);					// Store the file	  
			BIO_free(f);								// Cleanup
		}  
	}

  return ok;
}


/**
 * Do some testing.
 */
int doTest(TESTINFO testInfo, 
		   char *strCipher, 
		   RequestProtocol protocol, 
		   CERTINFO *certInfo, 
		   long *certRes, 
		   int *algBits, 
		   int *actuallyUsed, 
		   BOOL *bRenegotiatePossible,
		   BOOL *bSafeRenegotiation)
{
    SSL_CTX *ctx;
    SSL *ssl;
    BIO *sbio;
	SOCKET lhSocket;
	SOCKADDR_IN lSockAddr;
	int lConnect;
	int s_server_session_id_context = 1;
	int retval = 0;

	// 2010-03-24
	if(protocol == DTLS1)
	{
		lhSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	}
	else
	{
		lhSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	}

	if(lhSocket == INVALID_SOCKET)
	{
		printf("Invalid Socket %d", GetLastError());
		return -1;
	}

	memset(&lSockAddr,0, sizeof(lSockAddr));
	lSockAddr.sin_family = AF_INET;
	lSockAddr.sin_port = htons(testInfo.port);
	lSockAddr.sin_addr.s_addr = inet_addr(testInfo.strIPAdress);

	// Set the timeout for Read-operation.
	struct timeval tv;
	tv.tv_sec = testInfo.secondTimeout;
	tv.tv_usec = 0;

	// This timeout is important for our 3Com wlan gadget.
	if (setsockopt(lhSocket, SOL_SOCKET, SO_RCVTIMEO, (char *)&tv,  sizeof tv))
	{	
		return -1;
	}
    
    // Build our SSL context
    ctx=initCTX(protocol, testInfo.bUseWorkarounds, testInfo.bUseCiscoSpeshul);

	// Set the CA list (has to be called before we setup a connection)
    if(!(SSL_CTX_load_verify_locations(ctx, testInfo.strCACertPath, NULL)))
	{
		printf("[-] ERROR - Cannot find proper CA list file: %s", testInfo.strCACertPath);
		return -1;
	}


	// Never verify...
	SSL_CTX_set_verify(ctx,
                     SSL_VERIFY_NONE,
					 cert_verify_callback);
    

    // Set our cipher list 
	SSL_CTX_set_cipher_list(ctx, strCipher);
        
    // Connect the TCP socket
	lConnect = connect(lhSocket,(SOCKADDR *)&lSockAddr, sizeof(SOCKADDR_IN));
	if(lConnect != 0)
	{
		if(testInfo.bVerbose)
		{
			printf("[-] ERROR: Could not perform TCP-handshake. Return code: %d\n", lConnect);
		}

		return -1;
	}

	// Try to start TLS for the SMTP-service.
	if(testInfo.bStartTLS)
	{
		const int SBSIZE = 2048;
		char sockBuf[SBSIZE];
		char sockBufOut[SBSIZE];

		memset(sockBuf, 0, SBSIZE);
		recv(lhSocket, sockBuf, SBSIZE, 0);

		if (strncmp(sockBuf, "220", 3) != 0)
		{
			printf("[-] ERROR: SMTP-service not detected. The server responded: %s\n", sockBuf);
		}
		else 
		{
			if(testInfo.bVerbose)
			{
				printf("\n");
				printf("[+] Found SMTP-server. Now trying to start TLS...\n");
				printf("[+] %s\n", sockBuf);
			}

			// Send ehlo
			sprintf(sockBufOut, "EHLO %s\r\n", testInfo.strSMTPConnect);
			send(lhSocket, sockBufOut, strlen(sockBufOut), 0);


			// Receive OK
			memset(sockBuf, 0, SBSIZE);
			recv(lhSocket, sockBuf, SBSIZE, 0);

			if (strncmp(sockBuf, "250", 3) != 0)
			{
				printf("[-] ERROR: The SMTP service responded with error: %s\n", sockBuf);
			}
			else
			{
				if(testInfo.bVerbose)
				{
					printf("[+] %s\n", sockBuf);
				}
	
				// Send command
				strcpy(sockBufOut, "STARTTLS\r\n");
				send(lhSocket, sockBufOut, strlen(sockBufOut), 0);

				// Wait for response.
				memset(sockBuf, 0, SBSIZE);
				recv(lhSocket, sockBuf, SBSIZE, 0);


				if (strncmp(sockBuf, "220", 3) != 0)
				{
					printf("[-] ERROR: The SMTP-server does not want to start tls. Response: %s\n", sockBuf);
				}
			}
		}
	}


    // Connect the SSL socket 
    ssl=SSL_new(ctx);
    sbio=BIO_new_socket((int) lhSocket, BIO_NOCLOSE);
    SSL_set_bio(ssl, sbio, sbio);

	// Connect thru SSL.
	int conCode = SSL_connect(ssl);
    if(conCode <= 0)
	{			
		char buf[2048];
		int err = ERR_get_error();	// If the last byte of this is SSL_R_NO_CIPHERS_AVAILABLE (0xB5) 
									// then it is not valid for the specific protocol. (see ssl.h)
		
		ERR_error_string(err, buf);	// Retrieve the error string

		// If verbose we print some debug information.
		if(testInfo.bVerbose)
		{
			int retError = SSL_get_error(ssl, conCode);

			printf("\n[-] ERROR: SSL connect error: %d\n", retError);
			printf("[-] SSL Error: %d - %s\n", err, buf);				
		}


		// Cleanup
		closesocket(lhSocket);
		SSL_CTX_free(ctx);

		return CONNECT_FAIL;
	}



	retval = CONNECT_OK;								// Indicate success.		
	*actuallyUsed = SSL_get_cipher_bits(ssl, algBits);	// Check the number of bits.


	// Check certificate.
	if(testInfo.bCheckCert)
	{
		// Verify certificate result //X509_V_OK = 0
		long certResult = 0;
		
		// Check the certificate
		certInfo->certficate = NULL;
		certInfo->verResult = check_cert(ssl, testInfo.strHostname, certInfo->certficate, ctx, &certResult);	

		certInfo->certficate=SSL_get_peer_certificate(ssl);
		*certRes = certResult;	
	}

	// If we should test http.
	if(testInfo.bTestSend)
	{
		int req = sendRequest(ssl, testInfo);

		if(req == 1)
		{
			retval |= SEND_OK;

			int recv = receiveResponse(ssl, testInfo);

			if(recv == RECEIVE_OK)
			{
				retval |= RECEIVE_OK;
			}
		}
	}

	// Test for renegotiation.
	if(testInfo.bTestRenegotiation)
	{
		int renegotiateCode = SSL_renegotiate(ssl);

		if(renegotiateCode)
		{
			int handShakeCode = SSL_do_handshake(ssl);

			if(handShakeCode == 1)
			{
				*bRenegotiatePossible = TRUE;

				// Ok. We assume OpenSSL 1.0 from here on.
				long bSecureSupport = SSL_get_secure_renegotiation_support(ssl);

				if(bSecureSupport)
				{
					*bSafeRenegotiation = TRUE;
				}
			}
		}
	}



	// Cleanup.
	closesocket(lhSocket);
	SSL_CTX_free(ctx);


	return retval;
}




/**
 * Receive respnse
 */
int receiveResponse(SSL  *ssl, TESTINFO testInfo)
{
	#define BUFFER_SIZE 4096
	char buf[BUFFER_SIZE];
	int retval = RECEIVE_FAIL;
	int receivedLength = 0;
	BOOL receiveError = FALSE;

	if(testInfo.bPrintCommunication)
	{
		printf("\nReceived:\n\n");
	}

    // Now read the server's response, assuming that it's terminated by a close 
    while(1)
	{
		int r = SSL_read(ssl, buf, BUFFER_SIZE);
		int sslError = SSL_get_error(ssl, r);


		// If we did not receive anything then we have gotten a timeout.
		if(r <= 0)
		{
			if(receivedLength <= 0)
			{
				if(sslError == SSL_ERROR_SYSCALL)
				{
					int error = ERR_get_error();			

					if(error == 0)
					{
						if(r == -1)
						{
							printf("Error: The underlying BIO reported an I/O error\n");
						}
						else if(r == 0)
						{
							printf("Error: an EOF was observed that violates the protocol\n");
						}
					}
					else
					{
						printf("Error code: %d\n", error);
					}
				}
			}

			break;
		}

		if(testInfo.bPrintCommunication)
		{
			buf[r] = 0;
			printf("%s", buf);
		}

		

		receivedLength += r;

		if(sslError == SSL_ERROR_NONE)
		{
			continue;
		}
		else if(sslError == SSL_ERROR_WANT_READ)
		{
			continue;
		}
		else if(sslError == SSL_ERROR_ZERO_RETURN)
		{
			break;
		}
		else if(sslError == SSL_ERROR_SYSCALL)
		{
			break;
		}
		else
		{
			receiveError = TRUE;
			break;
		}
	}

	if(receivedLength > 0 && ! receiveError)
	{
		return RECEIVE_OK;
	}

	return RECEIVE_FAIL;
}

/**
 * send request
 */
int sendRequest(SSL *ssl, TESTINFO testInfo)
{
	char *request;
	char *nstr = NULL;
	int request_len = 0;
	BOOL bAllocated = FALSE;

	if(testInfo.strRequest == NULL)
	{
		request = new char[2048];
		bAllocated = TRUE;

		if(testInfo.testSIP)
		{
			
			char *REQUEST_TEMPLATE=
							"REGISTER sip:%s SIP/2.0\r\n"
							"Via: SIP/2.0/TLS test\r\n"
							"Max-Forwards: 70\r\n"
							"From: <sip:test@%s>\r\n"
							"To: <sip:test@%s>\r\n"
							"Call-ID: e0badd3c37c6462889498868cc1e2637\r\n"
							"CSeq: 1 REGISTER\r\n"
							//"Contact: <sip:192.168.1.114:9302;transport=tcp;ms-opaque=36a811c72c>;methods=\"INVITE, MESSAGE, INFO, OPTIONS, BYE, CANCEL, NOTIFY, ACK, REFER, BENOTIFY\";proxy=replace;+sip.instance=\"<urn:uuid:A2D11313-D4CC-596A-A604-D08CD3151892>\"\r\n"
							"User-Agent: UCCAPI/3.5.6907.0 OC/3.5.6907.0 (Microsoft Office Communicator 2007 R2)\r\n"
							"Supported: gruu-10, adhoclist, msrtc-event-categories\r\n"
							"Supported: ms-forking\r\n"
							"ms-keep-alive: UAC;hop-hop=yes\r\n"
							"Event: registration\r\n"
							"Content-Length: 0\r\n"
							"\r\n\0"
							;	

			// Create the request
			sprintf(request, REQUEST_TEMPLATE, testInfo.strSIPDomain, testInfo.strSIPDomain, testInfo.strSIPDomain);

		}
		else
		{
			char *REQUEST_TEMPLATE=
							"GET %s HTTP/1.0\r\n"
							"User-Agent: %s\r\n"
							"Host: %s\r\n\r\n\0";	

			// Create the request
			sprintf(request, REQUEST_TEMPLATE, testInfo.strHTTPPage, testInfo.strUseragent, testInfo.strGetHostname);
		}
	}
	else 
	{
		request = testInfo.strRequest;
        
		nstr = new char[strlen(request) + 1024];		
		strcpy(nstr, "");

		string str(request);
		int first = (int) str.find("\\r\\n", 0);		

		if(first > 0)
		{
			int last = 0;

			while(first > 0)
			{
				string strB = str.substr(last, first);
				char *strBC = (char *) strB.c_str();

				strcat(nstr, strBC);
				strcat(nstr, "\r\n");

				last = first + 4;
				first = (int) str.find("\\r\\n", last);
			}
			request = nstr;
		}		
	}

	// Figure out the request size and send it.
	request_len=(int) strlen(request);

	if(testInfo.bPrintCommunication)
	{
		printf("\n\nSending:\n%s\n", request);
	}

	int r = SSL_write(ssl, request, request_len);

	// Delete allocated space.
	if(bAllocated)
	{
		delete request;
	}

	if(nstr != NULL)
	{
		delete nstr;
	}

	if(request_len!=r)
	{
		return -1;
	}

	// Check for possible errors.
	switch(SSL_get_error(ssl, r))
	{
		case SSL_ERROR_NONE:
		{
			return 1;	
		}
		break;	
	}

	return -2;
}

